from __future__ import annotations

__all__ = [
    "roundas",
    "roundto",
    "atomic",
    "l10n",
]

import decimal

from linearmoney import cache
from linearmoney.data import CurrencyData, LocaleData


def _get_rounded_decimal(value: decimal.Decimal, places: int) -> decimal.Decimal:
    """Fixed-point rounding to user-supplied precision."""

    exponent = decimal.Decimal("10") ** -places
    rounded_value = value.quantize(exponent)
    return rounded_value


def _round_integral(value: decimal.Decimal, denomination: int) -> decimal.Decimal:
    """Integer rounding algorithm:

    Step 1: Round to integer.
    Step 2: Divide by denomination.
    Step 3: Round result to integer.
    Step 4: Multiply result by denomination.
    Step 5: Quantize result to zero decimal places for consistency.
    """

    _integral_quantizer = decimal.Decimal("1")
    _integral = value.quantize(_integral_quantizer)
    _rounded_integral = (_integral / denomination).quantize(
        _integral_quantizer
    ) * denomination
    return _rounded_integral.quantize(_integral_quantizer)


def _round_fractional(
    value: decimal.Decimal, denomination: int, places: int
) -> decimal.Decimal:
    """Fractional rounding algorithm:

    Step 1: Split the decimal value into its integral and fractional parts.
    Step 2: Truncate fractional part to the correct number of digits based on `places`.
    Step 3: Divide fractional part by denomination.
    Step 4: Round result to integer.
    Step 5: Multiply result by denomination.
    Step 6: Multiply integral part by 10 ** `places`.
    Step 7: Add the result of the fractional calculations to the result of the integral
    calculations.
    Step 8: Divide result of previous addition by 10 ** `places`.
    Step 9: Quantize final result to `places` for consistency.
    """

    if len(str(denomination)) > places:
        # `places` must be at least the number of digits in
        # `denomination.
        places = len(str(denomination))

    _exponent = decimal.Decimal("10") ** places
    _quantizer = decimal.Decimal("10") ** -places

    if value.to_integral_value() == value:
        # Integers rounded to specific places should round the integer value
        # based on the denomination and then quantize the result with the
        # correct number of zeros after the decimal point.
        return _round_integral(value, denomination).quantize(_quantizer)

    _decimal_split = str(value.normalize()).split(".")
    _integral_str, _fractional_str = _decimal_split[0], _decimal_split[1]
    _integral, _fractional = decimal.Decimal(_integral_str), decimal.Decimal(
        _fractional_str
    )
    _shift_width: int
    _fractional_digit_width = _fractional.adjusted() + 1
    if len(_fractional_str) > _fractional_digit_width:
        # Correct the shift width for any leading zeros that
        # were lost during string conversion.
        _shift_width = len(_fractional_str) - places
    else:
        _shift_width = _fractional_digit_width - places
    _truncated_fractional = _fractional.shift(-_shift_width)
    _rounded_fractional = (_truncated_fractional / denomination).quantize(
        decimal.Decimal("1")
    ) * denomination
    _rounded_value = ((_integral * _exponent) + _rounded_fractional) / _exponent
    return _rounded_value.quantize(_quantizer)


def _extract_fractions_data(currency: CurrencyData, cash: bool) -> tuple[int, int]:
    """Return the denomination and places values for the specified currency as a
    2-element tuple."""

    if cash:
        denomination = currency.data["cash_denomination"]
        places = currency.data["cash_places"]
    else:
        denomination = currency.data["denomination"]
        places = currency.data["places"]
    return denomination, places


@cache.cached()
def _get_rounded_value(
    value: decimal.Decimal, denomination: int, places: int
) -> decimal.Decimal:
    """Round `value` based on the provided `denomination` and `places`."""

    if denomination == 0 or denomination == 1:
        rounded_value = _get_rounded_decimal(value, places)
        return rounded_value
    else:
        if places <= 0:
            return _round_integral(value, denomination)
        else:
            return _round_fractional(value, denomination, places)


@cache.cached()
def roundas(
    amount: decimal.Decimal, currency: CurrencyData, cash: bool = False
) -> decimal.Decimal:
    """Round `amount` based on the denominational data defined by `currency`.

    If `cash` is True, round based on the *cash* denominations defined by `currency`.

    Example:

        >>> import linearmoney as lm
        >>> fv = lm.vector.forex({"base": "cad", "rates": {"usd": 1}})
        >>> av = lm.vector.asset(10.067777, "cad", lm.vector.space(fv))
        >>> val = lm.vector.evaluate(av, "cad", fv)
        >>> curr = lm.data.currency("cad")
        >>> lm.scalar.roundas(val, curr)
        Decimal('10.07')
        >>> lm.scalar.roundas(val, curr, cash=True)
        Decimal('10.05')
    """

    denomination, places = _extract_fractions_data(currency, cash=cash)

    return _get_rounded_value(amount, denomination, places)


@cache.cached()
def roundto(amount: decimal.Decimal, places: int) -> decimal.Decimal:
    """Round `amount` to a fixed number of decimal `places`.

    Example:

        >>> import linearmoney as lm
        >>> fv = lm.vector.forex({"base": "cad", "rates": {"usd": 1}})
        >>> av = lm.vector.asset(10.067777, "cad", lm.vector.space(fv))
        >>> val = lm.vector.evaluate(av, "cad", fv)
        >>> lm.scalar.roundto(val, 3)
        Decimal('10.068')
        >>> lm.scalar.roundto(val, 2)
        Decimal('10.07')
        >>> lm.scalar.roundto(val, 1)
        Decimal('10.1')
    """

    return _get_rounded_decimal(amount, places)


@cache.cached()
def atomic(amount: decimal.Decimal, currency: CurrencyData, cash: bool = False) -> int:
    """Return the integer value of `amount` in its smallest denomination as defined by
    the `currency` data.

    If `cash` is True, return the value in the smallest *cash* denomination defined by
    `currency`.

    Example:

        >>> import linearmoney as lm
        >>> fv = lm.vector.forex({"base": "cad", "rates": {"usd": 1}})
        >>> av = lm.vector.asset(10.07, "cad", lm.vector.space(fv))
        >>> val = lm.vector.evaluate(av, "cad", fv)
        >>> curr = lm.data.currency("cad")
        >>> lm.scalar.atomic(val, curr)
        1007
        >>> lm.scalar.atomic(val, curr, cash=True)
        1005
    """

    rounded_value = roundas(amount, currency, cash)
    return int("".join(str(rounded_value).split(".")))


@cache.cached()
def _format_grouping(rounded_value: decimal.Decimal, locale: LocaleData) -> str:
    """Format `rounded_value` based on the grouping and separator data of `locale`."""

    _sign = "positive" if rounded_value >= 0 else "negative"

    # TODO: Refactor this function for readability.
    str_value = str(rounded_value).strip("-").strip("+")
    decimal_separator = ""
    grouping_separator = locale.data["grouping_separator"]
    grouping = locale.data["_".join([_sign, "grouping"])]  # type: ignore[literal-required]
    decimal_split = str_value.split(".")
    decimal_part = ""
    if len(decimal_split) > 1:
        decimal_separator = locale.data["decimal_separator"]
        decimal_part = decimal_split[1]
    integer_part_string = decimal_split[0]
    formatted_number = ""
    if len(integer_part_string) <= grouping[0] or grouping == [-1]:
        formatted_number = f"{decimal_separator}".join(
            [integer_part_string, decimal_part]
        )
    else:
        integer_part = list(integer_part_string)
        integer_part.reverse()
        integer_split: list = []
        count = 0
        for i in reversed(grouping):
            if len(integer_part[count:]) < i:
                break
            integer_split.append(integer_part[count : count + i])
            count += i
        last_index = grouping[0]
        tmp_list = [
            integer_part[i : i + last_index]
            for i in range(count, len(integer_part), last_index)
        ]
        integer_split += tmp_list

        integer_split = ["".join(integer_split[i]) for i in range(len(integer_split))]

        reversed_integer_part = f"{grouping_separator}".join(integer_split)
        integer_part = list(reversed_integer_part)
        integer_part.reverse()
        formatted_integer_part = "".join(integer_part)
        formatted_number = f"{decimal_separator}".join(
            [formatted_integer_part, decimal_part]
        )
    return formatted_number


@cache.cached()
def l10n(
    amount: decimal.Decimal,
    currency: CurrencyData,
    locale: LocaleData,
    *,
    international: bool = False,
) -> str:
    """Localize `amount` to a currency-formatted string based on the `locale` and
    `currency` data.

    Args:
        amount:
            The decimal value to format. This will usually be a rounded decimal value
            resulting from a call to `roundas` or `roundto`, but it doesn't have to be.
        currency:
            The `linearmoney.data.CurrencyData`
            of the currency that the `amount` represents.
        locale:
            The `linearmoney.data.LocaleData` to use for formatting the result.
        international:
            Keyword-only argument. If False (default), then use the currency symbol for
            the target currency in the result and respect the `symbol_space` value of
            the `locale` data. If True, then use the international format with the
            ISO 4217 alpha currency code in place of the symbol and a space between
            the value and the symbol regardless of the value of the `symbol_space`
            data for the `locale`.

    Example:

        >>> import linearmoney as lm
        >>> fv = lm.vector.forex({"base": "cad", "rates": {"eur": 2}})  # 1 CAD -> 2 EUR
        >>> av = lm.vector.asset(10.067777, "cad", lm.vector.space(fv))
        >>> en_US = lm.data.locale("en", "us")  # English-United States
        >>> cad = lm.data.currency("cad")
        >>> val = lm.vector.evaluate(av, "cad", fv)
        >>> rounded_val = lm.scalar.roundas(val, cad)
        >>> lm.scalar.l10n(rounded_val, cad, en_US)
        'CA$10.07'
        >>> cash_rounded_val = lm.scalar.roundas(val, cad, cash=True)
        >>> lm.scalar.l10n(cash_rounded_val, cad, en_US)
        'CA$10.05'
        >>> lm.scalar.l10n(rounded_val, cad, en_US, international=True)
        'CAD 10.07'
        >>> fr_FR = lm.data.locale("fr", "fr")  # French-France
        >>> eur = lm.data.currency("eur")
        >>> val = lm.vector.evaluate(av, "eur", fv)
        >>> rounded_val = lm.scalar.roundas(val, eur)
        >>> lm.scalar.l10n(rounded_val, eur, fr_FR)
        '20,14€'
        >>> cash_rounded_val = lm.scalar.roundas(val, eur, cash=True)
        >>> lm.scalar.l10n(cash_rounded_val, eur, fr_FR)
        '20,14€'
        >>> lm.scalar.l10n(rounded_val, eur, fr_FR, international=True)
        '20,14 EUR'
    """

    # TODO: Refactor this function for readability.

    iso_code = currency.iso_code

    _sign = "positive" if amount >= 0 else "negative"

    lc_data = locale.data

    sign = lc_data["_".join([_sign, "sign"])]  # type: ignore[literal-required]
    symbol_before = lc_data["_".join([_sign, "symbol_before"])]  # type: ignore[literal-required]
    symbol_space = lc_data["_".join([_sign, "symbol_space"])]  # type: ignore[literal-required]
    sign_position = lc_data["_".join([_sign, "sign_position"])]  # type: ignore[literal-required]

    if not international:
        _currency_symbol = lc_data["currency_symbols"][iso_code]
        if _currency_symbol is None:
            international = True
        else:
            currency_symbol = _currency_symbol
    if international:
        currency_symbol = iso_code
        symbol_space = 1
        if sign_position == 1 and symbol_before:
            sign_position = 4
            symbol_space = 2

    join_list = []
    if sign_position == 3:
        if symbol_space == 2:
            currency_symbol = " ".join([sign, currency_symbol])
        else:
            currency_symbol = "".join([sign, currency_symbol])
    elif sign_position == 4:
        if symbol_space == 2:
            currency_symbol = " ".join([currency_symbol, sign])
        else:
            currency_symbol = "".join([currency_symbol, sign])

    if symbol_space == 1:
        join_list = [
            currency_symbol,
            " ",
            _format_grouping(amount, locale),
        ]
    else:
        join_list = [
            currency_symbol,
            _format_grouping(amount, locale),
        ]

    if not symbol_before:
        join_list.reverse()

    match sign_position:
        case 0:
            join_list.insert(0, "(")
            join_list.append(")")
        case 1:
            if symbol_space == 2 and not symbol_before:
                join_list.insert(0, "".join([sign, " "]))
            else:
                join_list.insert(0, sign)
        case 2:
            if symbol_space == 2 and symbol_before:
                join_list.append("".join([" ", sign]))
            else:
                join_list.append(sign)
    return "".join(join_list)



"""Functions for manipulating package resources."""

from __future__ import annotations


__all__: list[str] = [
    "get_package_resource",
]


import json
import importlib.resources
from os import PathLike
from typing import Literal, Any, TypeAlias

import linearmoney as lm

ResourceName: TypeAlias = (
    Literal["locales"]
    | Literal["currencies"]
    | Literal["supported_iso_codes"]
    | Literal["cldr_version"]
)


def _load_json_resource(res_path: str | PathLike) -> dict:
    with open(res_path, "r") as json_file:
        return json.load(json_file)


def get_package_resource(res_name: ResourceName) -> Any:
    """Load a JSON resource from the package's filesystem.

    Args:
        res_name:
            The name of the resource to load.
    Returns:
        The resource's data as a python structure.
    """

    fn = ".".join([res_name, "json"])
    with importlib.resources.as_file(
        importlib.resources.files(lm).joinpath(fn)
    ) as res_path:
        return _load_json_resource(res_path)



"""Exception classes raised by the linearmoney library."""

from __future__ import annotations

__all__ = [
    "SpaceError",
    "IntegrityError",
    "UnknownDataError",
    "CacheError",
]


class SpaceError(Exception):
    """An operation between vectors in two different spaces
    was attempted."""

    def __init__(self, msg: str) -> None:
        return super().__init__(msg)


class IntegrityError(Exception):
    """An operation that would result in breaking the
    mathematical invariants of the linear money model was attempted."""

    def __init__(self, msg: str) -> None:
        return super().__init__(msg)


class UnknownDataError(Exception):
    """An operation using unavailable data was attempted.

    For example, localizing to a locale that we don't have formatting data for,
    or rounding to a currency that we don't have denominational data for.
    """

    def __init__(self, msg: str) -> None:
        return super().__init__(msg)


class CacheError(Exception):
    """A caching operation failed."""

    def __init__(self, msg: str) -> None:
        super().__init__(msg)



from __future__ import annotations

__all__ = [
    "locale",
    "system_locale",
    "currency",
    "FormatType",
    "DataMap",
    "LocaleMap",
    "LocaleData",
    "CurrencyMap",
    "CurrencyData",
]

import copy
import enum
from collections.abc import MutableMapping, Mapping
from typing import TypedDict, Any, cast
from dataclasses import dataclass
import locale as posix_locale

from linearmoney import cache, resources
from linearmoney.mixins import ImmutableDeduplicationMixin, EqualityByHashMixin
from linearmoney.exceptions import UnknownDataError


class FormatType(enum.StrEnum):
    STANDARD = enum.auto()
    ACCOUNTING = enum.auto()


class DataMap(EqualityByHashMixin, ImmutableDeduplicationMixin, Mapping):
    """A read-only mapping."""

    __slots__ = ["_data", "_data_repr", "_hash"]

    def __init__(self, *args, **kwargs) -> None:
        self._data = dict(*args, **kwargs)
        self._data_repr = repr(self._data)
        self._hash = hash(self._data_repr)

    def __repr__(self) -> str:  # pragma: no cover
        _repr = getattr(self, "_data_repr", None)
        if _repr is None:
            return f"{self.__class__.__name__}({self._data})"
        else:
            return f"{self.__class__.__name__}({self._data_repr})"

    def __hash__(self) -> int:  # pragma: no cover
        return self._hash

    def __getitem__(self, key: str) -> Any:
        value = self._data[key]
        if isinstance(value, MutableMapping):
            # Ensure DataMap is read-only at all levels of nesting.
            return DataMap(value.items())
        return value

    def __iter__(self):  # pragma: no cover
        return self._data.__iter__()

    def __len__(self):  # pragma: no cover
        return self._data.__len__()


class LocaleMap(TypedDict):
    """Represents the structure of localization data for individual locales.

    TODO:
        Mark this as readonly once [PEP 705](https://peps.python.org/pep-0705/)
        is implemented.
        We currently have to cast a `DataMap` to `LocaleMap` whenever we construct this
        dict since we can't mark a `TypedDict` as read-only yet.
    """

    currency_symbols: MutableMapping[str, str]
    decimal_separator: str
    grouping_separator: str
    local_currency_code: str
    negative_grouping: tuple[int, ...]
    negative_sign: str
    negative_sign_position: int
    negative_symbol_before: bool
    negative_symbol_space: int
    positive_grouping: tuple[int, ...]
    positive_sign: str
    positive_sign_position: int
    positive_symbol_before: bool
    positive_symbol_space: int


@dataclass(eq=False, frozen=True)
class LocaleData(EqualityByHashMixin):
    """A [Datasource](/linearmoney/glossary.html#datasource) that provides formatting
    data for currency localization."""

    language: str
    region: str
    nformat: FormatType
    data: LocaleMap

    @property
    def id(self) -> tuple[str, str, FormatType]:
        """Identifies the locale, but not necessarily the unique data of the locale.

        Represents the positional arguments used to create this instance.
        See [Datasources](/linearmoney/glossary.html#datasource) in the glossary
        for more details.
        """

        return (self.language, self.region, self.nformat)

    @property
    def tag(self) -> str:
        """The [`Locale Tag`](/linearmoney/glossary.html#locale-tag) for this locale."""

        return "_".join([self.language, self.region])

    def __hash__(self) -> int:
        return hash((self.id, self.data))


_REQUIRED_LOCALE_KEYS = {
    "currency_symbols",
    "decimal_separator",
    "grouping_separator",
    "local_currency_code",
    "negative_grouping",
    "negative_sign",
    "negative_sign_position",
    "negative_symbol_before",
    "negative_symbol_space",
    "positive_grouping",
    "positive_sign",
    "positive_sign_position",
    "positive_symbol_before",
    "positive_symbol_space",
}


@cache.cached()
def _merge_locale_overrides(locales: LocaleMap, **overrides) -> LocaleMap:
    """Merge any values given by keyword arguments into the resulting locale data,
    overriding the values for the corresponding keys in `locales`."""

    new_locales = copy.deepcopy(locales)

    for i in overrides:
        if i in _REQUIRED_LOCALE_KEYS:
            if i == "currency_symbols":
                # Ensure currency codes used as keys are all upper case.
                symbols = {
                    k.upper(): v for k, v in overrides["currency_symbols"].items()
                }
                for k, v in symbols.items():
                    new_locales["currency_symbols"][k] = v
            else:
                new_locales[i] = overrides[i]  # type: ignore[literal-required]
    return new_locales


_fallback_locales = resources.get_package_resource("locales")


@cache.cached(size_multiplier=2)
def locale(
    language: str,
    region: str,
    nformat: FormatType = FormatType.STANDARD,
    **overrides,
) -> LocaleData:
    r"""Create a new `LocaleData` datasource for locale given by `language`, and
    `region`.

    Args:
        language:
            The language portion of the [locale tag](/linearmoney/glossary.html#locale-tag)
        region:
            The region (territory) portion of the
            [locale tag](/linearmoney/glossary.html#locale-tag)
        nformat:
            The number format to use. Should be a member of the `FormatType` string enum.
            The difference in format is generally only in negative numbers. E.g. -10 in
            'standard' vs. (10) in 'accounting'.
            If you aren't sure which format you need, you probably want 'standard'.
        **overrides:
            Formatting options to override the default formatting rules for the
            requested locale.
            The name of each kwarg is interpreted as a key in the `LocaleMap`
            returned as the `LocaleData.data` field, and its
            value will replace the value of that key in the output `DataMap`.
            E.g. `locale("en", "US", grouping_separator=';')` will return a `LocaleData`
            with formatting data for the 'en_US' locale, but using a semi-colon
            for grouping, so localizing '$1,000.00' using this data would result in
            '$1;000.00'.

            Values of a mapping type only override the keys provided, not the entire
            mapping. For example, if the formatting data for the requested locale
            contains `currency_symbols: {"EUR": "€", "USD": "$"}` by default, then
            providing `currency_symbols={"USD": "MySymbol"}` in the `locale` call will
            result in a `currency_symbols` value of `{"EUR": "€", "USD": "MySymbol"}`,
            not `{"USD": "MySymbol"}`.
    Returns:
        The newly parsed `LocaleData`.
    Raises:
        `linearmoney.exceptions.UnknownDataError`:
            If the cldr-json data that linearmoney utilizes does not have formatting
            data for the [locale tag](/linearmoney/glossary.html#locale-tag) built from the
            `language` and `region` arguments.
        ValueError:
            If the `nformat` is not one of the supported string constants.
            This should also be caught by static type checkers since the argument is
            typed with a Literal type.
    """

    if not language.islower():
        language = language.lower()

    if not region.isupper():
        region = region.upper()

    locales: LocaleMap

    if nformat in _fallback_locales:
        locale_string = "_".join([language, region])
        try:
            locales = _fallback_locales[nformat][locale_string]
        except KeyError as e:
            raise UnknownDataError(
                f"invalid value for `language` {language} or `region` {region}. \
Locale data not available under number format {nformat}"
            ) from e

    else:
        raise ValueError(
            f"invalid value for `nformat`. Expected either 'standard' or \
'accounting' got {nformat}"
        )

    if overrides:
        # We cast to satisfy mypy since we can't use read-only TypedDict yet.
        # Remove this cast once pep 705 is implemented.
        return LocaleData(
            language,
            region,
            nformat,
            data=cast(
                LocaleMap, DataMap(_merge_locale_overrides(locales, **overrides))
            ),
        )
    else:
        # We cast to satisfy mypy since we can't use read-only TypedDict yet.
        # Remove this cast once pep 705 is implemented.
        return LocaleData(
            language, region, nformat, data=cast(LocaleMap, DataMap(locales))
        )


# Set up system locale.
if None in posix_locale.getlocale(posix_locale.LC_MONETARY):
    # Set the locale of the python session to the running system locale.
    posix_locale.setlocale(posix_locale.LC_ALL, "")


def system_locale() -> LocaleData:
    """The `LocaleData` of the current POSIX locale of the running
    Python process.

    This is useful for local applications such as commandline scripts
    and desktop apps since those applications usually don't need to support
    multiple locales, but the developer doesn't always know the locale of the
    end user's environment.

    Prior to version 0.1.2, there was a bug in this function that would cause
    a crash when the system locale was set to the default C/POSIX locale.
    Version 0.1.2 and later fix this by interpreting the default C/POSIX locale
    as `en_US`. See
    [#14](https://github.com/GrammAcc/linearmoney/issues/14).
    """

    system_locale_string: str | None = posix_locale.getlocale()[0]
    assert (
        system_locale_string is not None
    ), "We init the system locale above, so it should not be None."
    if system_locale_string.upper() == "C" or system_locale_string.upper() == "POSIX":
        system_locale_string = "en_US"
    language, region = system_locale_string.split("_")

    return locale(language, region)


class CurrencyMap(TypedDict):
    """Represents the structure of denominational/rounding data for currencies.

    TODO:
        Mark this as readonly once [PEP 705](https://peps.python.org/pep-0705/)
        is implemented.
        We currently have to cast a `DataMap` to `CurrencyMap` whenever we construct this
        dict since we can't mark a `TypedDict` as read-only yet.
    """

    places: int
    cash_places: int
    denomination: int
    cash_denomination: int


@dataclass(eq=False, frozen=True)
class CurrencyData(EqualityByHashMixin):
    """A [Datasource](/linearmoney/glossary.html#datasource) that provides denominational
    data for currency rounding."""

    iso_code: str
    data: CurrencyMap

    @property
    def id(self) -> tuple[str]:
        """Identifies the currency, but not necessarily the unique data of the currency.

        Represents the positional arguments used to create this instance.
        See [Datasources](/linearmoney/glossary.html#datasource) in the glossary
        for more details.
        """

        return (self.iso_code,)

    def __hash__(self) -> int:
        return hash((self.id, self.data))


_REQUIRED_CURRENCY_KEYS = {
    "places",
    "cash_places",
    "denomination",
    "cash_denomination",
}


@cache.cached()
def _merge_currency_overrides(currencies: CurrencyMap, **overrides) -> CurrencyMap:
    """Merge any values given by keyword arguments into the resulting currencies data
    overriding the values for the corresponding keys in `currencies`."""

    new_currencies = copy.deepcopy(currencies)

    for i in _REQUIRED_CURRENCY_KEYS:
        if i in overrides:
            new_currencies[i] = overrides[i]  # type: ignore[literal-required]

    return new_currencies


_fallback_currencies = resources.get_package_resource("currencies")
_supported_iso_codes = set(resources.get_package_resource("supported_iso_codes"))


@cache.cached()
def currency(iso_code: str, **overrides) -> CurrencyData:
    r"""Create a new `CurrencyData` datasource based on `iso_code` and `**overrides`

    Args:
        iso_code:
            The ISO 4217 alpha code of the currency to parse denominational data for.
        **overrides:
            Denominational options to override the default rounding rules for the
            requested currency.
            The name of each kwarg is interpreted as a key in the `CurrencyMap`
            returned as the `CurrencyData.data` field, and
            its value will replace the value of that key in the output `DataMap`.
            E.g. `currency("USD", denomination=5)` will return a `CurrencyData`
            with denominational data for the 'USD' currency, but rounding to the
            nearest $0.05, so rounding $10.068 using this data would result in
            $10.05, not $10.07.
    Returns:
        The newly parsed `CurrencyData`.
    Raises:
        `linearmoney.exceptions.UnknownDataError`:
            If the cldr-json data does not have denominational data for `iso_code`
            and not all of the fields of `CurrencyMap` are provided as **overrides.
    """

    if not iso_code.isupper():
        iso_code = iso_code.upper()

    if iso_code not in _supported_iso_codes:
        if (
            "denomination" not in overrides
            or "places" not in overrides
            or "cash_denomination" not in overrides
            or "cash_places" not in overrides
        ):
            raise UnknownDataError(
                f"Rounding data for currency {iso_code} not found.\
You must provide all rounding data for unknown currency."
            )

    currencies: CurrencyMap
    if iso_code in _fallback_currencies:
        currencies = _fallback_currencies[iso_code]
    else:
        currencies = _fallback_currencies["DEFAULT"]

    if overrides:
        # We cast to satisfy mypy since we can't use read-only TypedDict yet.
        # Remove this cast once pep 705 is implemented.
        return CurrencyData(
            iso_code,
            data=cast(
                CurrencyMap, DataMap(_merge_currency_overrides(currencies, **overrides))
            ),
        )
    else:
        # We cast to satisfy mypy since we can't use read-only TypedDict yet.
        # Remove this cast once pep 705 is implemented.
        return CurrencyData(iso_code, data=cast(CurrencyMap, DataMap(currencies)))



"""Helper functions used internally by the linearmoney package."""

from __future__ import annotations


__all__: list[str] = [
    "coerce_decimal",
    "get_supported_numeric",
    "is_supported_numeric",
    "perf",
]

import decimal
import time
import logging
from functools import singledispatch
from typing import Callable, TypeVar, ParamSpec

perf_logger = logging.getLogger("performance")
file_handler = logging.FileHandler(
    "performance.log",
    mode="w",
    encoding="utf-8",
    delay=True,
)
perf_logger.setLevel(logging.DEBUG)
file_handler.setLevel(logging.DEBUG)
perf_logger.addHandler(file_handler)

T = TypeVar("T")
P = ParamSpec("P")
iterations = [10, 100, 1_000, 10_000, 100_000, 1_000_000]


def perf(func: Callable[P, T], *args, **kwargs) -> None:
    for i in iterations:
        start = time.perf_counter_ns()
        for j in range(i):
            func(*args, **kwargs)
        end = time.perf_counter_ns()
        elapsed = end - start
        perf_logger.debug(
            "".join(
                [
                    func.__qualname__,
                    ": Iterations - ",
                    str(i),
                    ", Runtime - ",
                    str(elapsed),
                ]
            )
        )


@singledispatch
def coerce_decimal(number) -> decimal.Decimal:
    """Generic function to convert registered types to `decimal.Decimal` instances.

    New types can be registered with `coerce_decimal.register` at runtime, but this will
    likely break static type checking, so it is not recommended.
    See the documentation for
    [`functools.singledispatch`](https://docs.python.org/3/library/functools.html#functools.singledispatch)
    for details on usage.

    Default registered types are: `int`, `float`, `decimal.Decimal`, and
    `str`.

    Notes:
        - `str` is registered to allow valid string inputs to the `decimal.Decimal`
        constructor to be more easily checked as valid numeric input, but strings that
        are invalid as input to the `decimal.Decimal` constructor will still raise a
        `NotImplementedError` unless the user registers a new function for
        the `str` type to implement the conversion.
    """

    raise NotImplementedError


@coerce_decimal.register
def _force_int_to_decimal(
    number: int,
) -> decimal.Decimal:
    return decimal.Decimal(number)


@coerce_decimal.register
def _force_float_to_decimal(number: float) -> decimal.Decimal:
    return decimal.Decimal(str(number))


@coerce_decimal.register
def _force_decimal_to_decimal(number: decimal.Decimal) -> decimal.Decimal:
    return number


@coerce_decimal.register
def _force_numeric_string_to_decimal(number: str) -> decimal.Decimal:
    try:
        return decimal.Decimal(number)
    except decimal.InvalidOperation:
        raise NotImplementedError


def get_supported_numeric() -> tuple[type, ...]:
    """A tuple of all types that can be safely converted to `decimal.Decimal` by the
    linearmoney package at runtime.

    Corresponds to the types registered to the `coerce_decimal` generic
    function.
    """

    supported_classes = dict(coerce_decimal.registry)
    del supported_classes[object]
    return tuple(supported_classes)


def is_supported_numeric(value: object) -> bool:
    """Determines if a variable is of a type that can be safely converted to Decimal
    at runtime through the `coerce_decimal` generic function."""

    return type(value) in get_supported_numeric()



from __future__ import annotations

__all__: list[str] = [
    "ImmutableDeduplicationMixin",
    "EqualityByHashMixin",
]


from abc import ABCMeta, abstractmethod
from typing import ClassVar, Self
from collections.abc import Hashable, Sequence


class ImmutableDeduplicationMixin:
    """A mixin class that provides common dunder method implementations for
    logically immutable objects.

    Provides recursive implementations of `__getstate__` and
    `__setstate__` based on the `__slots__` declared on the class and all of its
    ancestors in order to fully support pickling without having
    to implement the methods in every class and provides implementations of `__copy__`
    and `__deepcopy__` that simply return self since copying makes no difference for
    immutable objects.
    """

    __slots__: ClassVar[Sequence[str]] = []

    def __getstate__(self) -> dict:
        state = {}
        for i in self.__class__.mro():
            sl = getattr(i, "__slots__", None)
            if sl is not None:
                for j in sl:
                    obj = getattr(self, j, None)
                    state[j] = obj
        return state

    def __setstate__(self, state: dict) -> None:
        for k, v in state.items():
            setattr(self, k, v)

    def __copy__(self) -> Self:
        # Just return self since we're immutable.
        return self

    def __deepcopy__(self, memo: dict = {}) -> Self:
        # Just return self since we're immutable.
        return self


class EqualityByHashMixin(metaclass=ABCMeta):
    """Mixin class that adds equality comparison support for user-defined hashable
    classes.

    Adding this mixin to a class will add a basic implementation of
    `__eq__` and `__ne__` which will compare objects for equality by hash
    instead of `id`. Only user defined classes that override
    `object.__hash__` should inherit from this mixin.

    This mixin does not add any of the ordered comparison operators
    (<, >, <=, >=), only equality comparison is added.
    """

    __slots__: ClassVar[Sequence[str]] = []

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Hashable):
            return hash(self) == hash(other)
        else:
            return NotImplemented  # pragma: no cover

    def __ne__(self, other: object) -> bool:
        if isinstance(other, Hashable):
            return hash(self) != hash(other)
        else:
            return NotImplemented  # pragma: no cover

    @abstractmethod
    def __hash__(self) -> int:
        raise NotImplementedError



"""Functions and types for doing monetary calculations in linear algebra."""

from __future__ import annotations

__all__: list[str] = [
    "asset",
    "forex",
    "basis_vector",
    "space",
    "gamma",
    "dot",
    "evaluate",
    "convert",
    "store",
    "restore",
    "MoneyVector",
    "ForexVector",
    "CurrencySpace",
    "RatesDict",
]

import decimal
import copy
from typing import TypeAlias, Self, TypedDict, TypeVar
from collections.abc import Iterator

from linearmoney import cache, _utils
from linearmoney.mixins import ImmutableDeduplicationMixin, EqualityByHashMixin
from linearmoney.exceptions import SpaceError, IntegrityError

DecimalVector: TypeAlias = tuple[decimal.Decimal, ...]

_ZERO = decimal.Decimal(0)
_ONE = decimal.Decimal(1)

# TypeVar needed for typechecking subclasses returned in place of base class.
V = TypeVar("V", bound="MoneyVector")


class CurrencySpace(ImmutableDeduplicationMixin, EqualityByHashMixin):
    """Represents the currency space of a `MoneyVector`."""

    __slots__ = ["_axes", "_currencies", "_hash"]

    def __init__(self, axes: tuple[str, ...]) -> None:
        """
        Args:
            axes:
                A tuple of ISO 4217 currency codes representing the euclidean axes of the
                vector space.
        """

        self._axes = axes
        self._currencies = set(axes)
        self._hash = hash((axes, repr(self._currencies)))

    def __repr__(self) -> str:
        return "".join(["CurrencySpace", str(self.axes)])

    @property
    def axes(self) -> tuple[str, ...]:
        """The tuple of ISO 4217 currency codes representing the axes of this
        currency space."""

        return self._axes

    @property
    def currencies(self) -> set[str]:
        """The axes of this currency space as a python `Set`.

        Useful for faster membership testing of currencies.

        Example:
            >>> import linearmoney as lm
            >>> fo = lm.vector.forex({"base": "EUR", "rates": {"USD": 0.4}})
            >>> sp = lm.vector.space(fo)
            >>> "USD" in sp.currencies  # Is USD part of the currency space?
            True
        """

        return self._currencies

    def __hash__(self) -> int:
        return self._hash


class MoneyVector(ImmutableDeduplicationMixin, EqualityByHashMixin):
    """Base class for all money vectors.

    `MoneyVector`s are iterables that behave like tuples of `decimal.Decimal`s.

    Implements the basic arithmetic operations for vector math.
    """

    __slots__ = ["_vector", "_v_repr", "_axes", "_hash"]

    def __init__(self, decimal_vector: DecimalVector, axes: tuple[str, ...]) -> None:
        """
        Args:
            decimal_vector:
                The actual componenents of the vector.
            axes:
                The ISO 4217 currency codes that correspond to the axes of the vector's
                currency space.
        Raises:
            `linearmoney.exceptions.SpaceError`:
                If `decimal_vector` and `axes` are not the same length.
        """

        if len(decimal_vector) != len(axes):
            raise SpaceError("Must have the same number of components as axes.")
        self._vector = decimal_vector
        self._v_repr = str(tuple([str(component) for component in decimal_vector]))
        self._axes = axes
        self._hash = hash((self._vector, self._axes))

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        # Preserve precision of Decimals
        return f"{self.__class__.__name__}{self._v_repr}"

    def __hash__(self) -> int:
        return self._hash

    def __getitem__(self, idx: int) -> decimal.Decimal:
        return self._vector[idx]

    def __iter__(self) -> Iterator:
        return self._vector.__iter__()

    def __len__(self) -> int:
        return len(self._vector)

    @cache.cached()
    def __add__(self, other: MoneyVector) -> Self:
        # Raises `SpaceError` if the two vectors are not part of the same currency
        # space.

        if not isinstance(other, MoneyVector):
            return NotImplemented

        if self.axes != other.axes:
            raise SpaceError("MoneyVectors must be in the same space.")

        new_vector = tuple([self[i] + other[i] for i in range(self.dim)])
        return self.__class__(new_vector, self.axes)

    @cache.cached()
    def __radd__(self, other: MoneyVector) -> Self:
        if other == 0:
            return self
        return self.__add__(other)

    @cache.cached()
    def __sub__(self, other: MoneyVector) -> Self:
        # Raises `SpaceError` if the two vectors are not part of the same currency
        # space.

        if not isinstance(other, MoneyVector):
            return NotImplemented

        if self.axes != other.axes:
            raise SpaceError("MoneyVectors must be in the same space.")

        new_vector = tuple([self[i] - other[i] for i in range(self.dim)])
        return self.__class__(new_vector, self.axes)

    @cache.cached()
    def __rsub__(self, other: V) -> V:
        if not isinstance(other, MoneyVector):
            return NotImplemented
        return other.__sub__(self)  # pragma: no cover

    @cache.cached()
    def __mul__(self, scalar: decimal.Decimal | int | float | str) -> Self:
        try:
            scalar = _utils.coerce_decimal(scalar)
        except NotImplementedError:
            return NotImplemented
        return self.__class__(tuple([scalar * i for i in self]), self.axes)

    @cache.cached()
    def __rmul__(self, scalar: decimal.Decimal | int | float | str) -> Self:
        return self.__mul__(scalar)

    @cache.cached()
    def __truediv__(self, scalar: decimal.Decimal | int | float | str) -> Self:
        try:
            scalar = _utils.coerce_decimal(scalar)
        except NotImplementedError:
            return NotImplemented
        return self * (decimal.Decimal(1) / scalar)

    @cache.cached()
    def __pos__(self) -> Self:
        return copy.deepcopy(self)

    @cache.cached()
    def __neg__(self) -> Self:
        return self.__class__(tuple([-i for i in self]), self.axes)

    @property
    @cache.cached()
    def dim(self) -> int:
        """The dimension of this `MunnyVector`."""

        return len(self)

    @property
    @cache.cached()
    def axes(self) -> tuple[str, ...]:
        """Tuple of the ISO 4217 alpha currency codes representing the axes of this
        `MunnyVector`."""

        return self._axes


class ForexVector(MoneyVector):
    """A [forex vector](/linearmoney/glossary.html#forex-vector) in the
    [linear money model](/linear_money_model.html)."""

    @cache.cached()
    def __neg__(self) -> Self:
        raise (IntegrityError("Forex vectors can not have negative-valued components."))


@cache.cached()
def basis_vector(currency_space: CurrencySpace, axis: str) -> MoneyVector:
    """Return the Euclidean basis vector corresponding to `axis` in `currency_space`.

    Raises:
        `linearmoney.exceptions.SpaceError`:
            If `axis` is not part of `currency_space`.
    """

    axis = axis.upper()

    if axis not in currency_space.currencies:
        raise SpaceError(
            f"Currency {axis} is not part of currency space {currency_space}"
        )
    _vector = tuple([_ONE if i == axis else _ZERO for i in currency_space.axes])
    return MoneyVector(_vector, currency_space.axes)


@cache.cached()
def dot(vec1: MoneyVector, vec2: MoneyVector) -> decimal.Decimal:
    """Calculate and return the dot product of vectors `vec1` and `vec2`.

    Raises:
        `linearmoney.exceptions.SpaceError`:
            If the two vectors are not part of the same currency space.
    """

    if vec1.axes != vec2.axes:
        raise SpaceError("MoneyVectors must be in the same space.")

    product = decimal.Decimal(
        str(sum([vec1[idx] * vec2[idx] for idx in range(vec1.dim)]))
    )
    return product


@cache.cached()
def space(vec: MoneyVector) -> CurrencySpace:
    """Create and return a new `CurrencySpace` representing the space that `vec`
    belongs to."""

    return CurrencySpace(vec.axes)


@cache.cached(size_multiplier=16)
def asset(
    amount: int | float | decimal.Decimal, iso_code: str, currency_space: CurrencySpace
) -> MoneyVector:
    """Create a new [*Asset Vector*](/linearmoney/glossary.html#asset-vector) in `currency_space`
    with `amount` of `iso_code` as it's only non-zero component.

    Raises:
        `linearmoney.exceptions.SpaceError`:
            If `iso_code` is not part of `currency_space`.
    """

    iso_code = iso_code.upper()

    if iso_code not in currency_space.currencies:
        raise SpaceError(f"{iso_code} is not part of {currency_space}")

    try:
        _amount = _utils.coerce_decimal(amount)
    except NotImplementedError:
        raise TypeError(f"Unsupported type for argument `amount`: {type(amount)}")

    _vector = tuple([_amount if i == iso_code else _ZERO for i in currency_space.axes])
    return MoneyVector(_vector, currency_space.axes)


_DecimalRates: TypeAlias = dict[str, decimal.Decimal]
_NumericRates: TypeAlias = dict[str, int | float | decimal.Decimal]


class RatesDict(TypedDict):
    """A dictionary representing forex rates ***from*** `base` ***to*** each of the
    keys in `rates`.

    The key names of this dict are chosen to match the most common structure returned
    by public HTTP APIs for forex data since that is how most applications
    will source their forex rates, so this allows passing the results of e.g.
    `json.load(api_response)` directly to any function that accepts this type.
    """

    base: str
    rates: _NumericRates


@cache.cached()
def _coerce_decimal_rates(numeric_rates: _NumericRates) -> _DecimalRates:
    """Coerce the values of `numeric_rates` to `decimal.Decimal` and force all keys
    to uppercase."""

    return {k.upper(): _utils.coerce_decimal(v) for k, v in numeric_rates.items()}


@cache.cached()
def _merge_rates(base_rates: _DecimalRates, overrides: _DecimalRates) -> _DecimalRates:
    """Override values in `base_rates` with corresponding values in `overrides`.
    Any keys in `overrides` not present in `base_rates` will be merged into `base_rates
    as well."""

    r = {k: v for k, v in base_rates.items()}
    r.update(overrides)
    return r


@cache.cached()
def _invert_rates(rates: _DecimalRates) -> _DecimalRates:
    """Invert the forex rates given by `rates`.

    Input forex rates are assumed to represent the rates from
    Quote Currency -> All Currencies. This function inverts them to represent the
    rates from All Currencies -> Quote Currency.

    E.g. (EUR -> EUR, EUR -> JPY, EUR -> USD) becomes
    (EUR -> EUR, JPY -> EUR, USD -> EUR).
    """

    return {k: (_ONE / v) for k, v in rates.items()}


@cache.cached()
def forex(
    forex_rates: RatesDict, **overrides: int | float | decimal.Decimal
) -> ForexVector:
    r"""Construct a new `ForexVector` suitable for use in the `gamma` function.

    Args:
        forex_rates:
            A mapping where each of the values in `forex_rates["rates"]` specifies
            the rate ***from*** `forex_rates["base"]` ***to*** the value's
            corresponding key.
            `forex_rates["base"]` and all keys in `forex_rates["rates"]` are treated as
            case-insensitive, and all numeric rate values are forced to
            `decimal.Decimal`, so `{"base": "eur", "rates": {"usd": 0.1}}` and
            `{"base": "EUR", "rates": {"USD": decimal.Decimal("0.1")}}` are treated as
            equivalent.
            If the `rates` key of `forex_rates` is an empty dictionary, then
            we create a forex vector with only one rate, which is `base -> base = 1`.
            This makes it clear that a forex vector is intended to force
            [single-currency](/user_guide/recipes.html#single-currency-application) calculations.
        \*\*overrides:
            Additional kwargs interpreted as ISO 4217 currency codes that should be
            manually overriden with the value provided or added to the set of
            exchange rates if the code is not already present in
            `forex_rates["rates"]`.
            Any rates passed as kwargs are interpreted in
            the same way as if they had been included in the `forex_rates["rates"]`
            nested mapping, so they should represent the exchange rates ***from***
            `forex_rates["base"]` ***to*** the currency represented by the name of
            the kwarg. All kwargs are case-insensitive, so passing `usd=0.1` will
            result in the rate
            `forex_rates["base"] -> "USD" = decimal.Decimal("0.1")`.
    Returns:
        The newly created `ForexVector`.

        Its [`axes`](#MoneyVector.axes) tuple will be the set of all
        currencies provided via `forex_rates["rates"]` and `overrides` along with
        the `forex_rates["base"]`.
    Raises:
        `linearmoney.exceptions.IntegrityError`:
            If the `forex_rates["rates"]` dictionary contains any values that are less
            than or equal to 0.
    """

    _rates: _DecimalRates = _coerce_decimal_rates(forex_rates["rates"])
    _iso_code: str = forex_rates["base"].upper()

    if overrides:
        _rates = _merge_rates(_rates, _coerce_decimal_rates(overrides))

    for r in _rates.values():
        if r <= 0:
            raise IntegrityError("All components of a ForexVector must be > 0.")

    # We copy here to prevent mutating rates returned from the cache
    _rates = {k: v for k, v in _rates.items()}
    _rates[_iso_code] = decimal.Decimal("1.0")

    _rates = _invert_rates(_rates)
    _sorted_rates = dict(sorted(_rates.items()))
    return ForexVector(tuple(_sorted_rates.values()), tuple(_sorted_rates))


@cache.cached()
def _round_forex(vec: ForexVector, quantizer: decimal.Decimal) -> ForexVector:
    """Return a new `ForexVector` with all components of `vec` rounded based on
    `quantizer`."""

    return ForexVector(tuple([r.quantize(quantizer) for r in vec]), vec.axes)


@cache.cached()
def gamma(r: ForexVector, iso_code: str, decimal_places: int = 17) -> ForexVector:
    """Return a new `ForexVector` representing the rates ***from*** all different
    currencies in the currency space of `r` ***to*** `iso_code`.

    Args:
        r:
            The forex vector to calculate the gamma vector from.
            Corresponds to the `r` of the [linear money model's](/linear_money_model.html)
            gamma function.
        iso_code:
            The ISO 4217 currency code that the resulting gamma vector will represent
            the rates ***to***.
        decimal_places:
            The number of decimal places the rates in the resulting gamma vector will
            use.
            This should be large enough to support the precise exchange rates between
            the most valuable and least valuable currencies in the currency space.
            This is NOT the decimal *precision*. It specifies the exact number of
            digits *after* the decimal separator, not the total digits of the entire
            number.
    Raises:
        `linearmoney.exceptions.SpaceError`:
            If `iso_code` is not part of the currency space of `r`.
    """

    _quantizer = decimal.Decimal("1.0") ** decimal_places

    iso_code = iso_code.upper()

    _space = space(r)
    if iso_code not in _space.currencies:
        raise SpaceError(f"{iso_code} is not an axis in currency space {_space}.")

    ek = basis_vector(_space, iso_code)
    conversion_factor = _ONE / dot(r, ek)
    new_rates_vector = r * conversion_factor
    return _round_forex(new_rates_vector, _quantizer)


_EVALUATION_QUANTIZER = decimal.Decimal("10") ** decimal.Decimal("-12")


@cache.cached()
def evaluate(
    asset_vec: MoneyVector, iso_code: str, forex_vec: ForexVector
) -> decimal.Decimal:
    """[Evaluate](/linearmoney/glossary.html#evaluation) `asset_vec` to `iso_code` using rates
    defined by `forex_vec`."""

    result = dot(asset_vec, gamma(forex_vec, iso_code))
    # Eliminate intermediate rounding errors.
    rounded_result = result.quantize(_EVALUATION_QUANTIZER)
    # Normalize to improve consistency and improve testability.
    normalized_result = rounded_result.normalize()
    return normalized_result


@cache.cached()
def convert(
    asset_vec: MoneyVector, iso_code: str, forex_vec: ForexVector
) -> MoneyVector:
    """[Convert](/linearmoney/glossary.html#conversion) `asset_vec` to `iso_code` using the rates
    defined by `forex_vec`.

    Returns a [*Rudimentary Asset*](/linearmoney/glossary.html#rudimentary-asset).
    """

    return evaluate(asset_vec, iso_code, forex_vec) * basis_vector(
        space(asset_vec), iso_code
    )


@cache.cached()
def store(vec: MoneyVector) -> str:
    """Serialize `vec` to a string that can be used to recreate the exact same
    vector."""

    component_list = [
        ";".join([k, str(v.normalize())]) for k, v in zip(vec.axes, vec, strict=True)
    ]
    return ":".join(component_list)


@cache.cached()
def restore(serial_str: str) -> MoneyVector:
    """Recreate a `MoneyVector` from a serialized string.

    Args:
        serial_str:
            A serialized string representation of a `MoneyVector`.
            Should be the same format as returned by the `store` function.
    Returns:
        `MoneyVector`

        The reconstructed vector.
        Should be exactly the same as the vector that was `store`d.
    Raises:
        ValueError:
            If the `serial_str` has an invalid format.
    """

    components = serial_str.split(":")
    new_decimal_vector_list = []
    new_axes_list = []
    for i in components:
        component = i.split(";")
        if len(component) != 2:
            raise ValueError(f"Invalid serial string format: {serial_str}")
        new_decimal_vector_list.append(decimal.Decimal(component[1]))
        new_axes_list.append(component[0])
    new_decimal_vector = tuple(new_decimal_vector_list)
    new_axes = tuple(new_axes_list)
    return MoneyVector(new_decimal_vector, new_axes)



"""This module provides access to the custom least recently used caching system used
by functions in the linearmoney package."""

from __future__ import annotations

__all__ = [
    "cached",
    "enable",
    "is_enabled",
    "head",
    "tail",
    "invalidate",
    "size",
    "max_size",
    "get_base_size",
    "set_base_size",
]

import logging

import decimal
import threading

import functools
from collections import OrderedDict
from collections.abc import Hashable, Callable, Mapping
from typing import (
    TypeVar,
    ParamSpec,
    SupportsInt,
    Any,
)

from linearmoney.exceptions import CacheError

logger = logging.getLogger(__name__)

_thread_local_data = threading.local()


class _LRUFuncCache(OrderedDict):
    """OrderedDict representing the lru cache of a specific function."""

    __slots__ = [
        "_funcname",
        "_size_multiplier",
    ]

    def __init__(
        self,
        from_store: Mapping = {},
        *,
        funcname: str,
        size_multiplier: int | float = 1,
    ) -> None:
        super().__init__(from_store)
        self._funcname = funcname
        self._size_multiplier = size_multiplier

    @property
    def max_size(self) -> int:
        """**Read-only**: The maximum number of cache
        entries for this `LRUStore`.

        Calculated: `base_size` * `size_multiplier` = `max_size`
        """

        return int(get_base_size() * self.size_multiplier)

    @property
    def size_multiplier(self) -> int | float:
        """**Read-only**: Contributes to dynamic `max_size`."""

        return self._size_multiplier

    @property
    def size(self) -> int:
        """Read-only: the current number of values cached on this instance."""

        return len(self)

    @property
    def head(self) -> Any:
        """Read-only: the cached value at the current *head* (least recently used)
        position."""

        return self[next(iter(self))]

    @property
    def tail(self) -> Any:
        """Read-only: the cached value at the current *tail* (most recently used)
        position."""

        return self[next(iter(reversed(self)))]

    def is_cached(self, cache_key: tuple) -> bool:
        return cache_key in self

    def write(self, cache_key: tuple, value: Any) -> None:
        """Cache a new value to this store.

        Args:
            cache_key
                The tuple key used to lookup the value within this funccache.
            value
                The value to be stored in this funccache.
        """

        self[cache_key] = value
        self.move_to_end(cache_key)
        if self._overfull():
            logger.warning(
                f"{self._funcname}: \
cache full on write(key={cache_key}, value={value})."
            )
            self._remove_head()

    def _overfull(self) -> bool:
        return len(self) > self.max_size

    def _remove_head(self) -> None:
        del self[next(iter(self))]

    def read(self, cache_key: tuple) -> Any:
        """Fetch a cached value from this store.

        Args:
            cache_key
                The tuple key used to lookup the value within this funccache.

        Returns:
            Any

            The value stored by `cache_key` in this store.
        """

        read_value = self[cache_key]
        self.move_to_end(cache_key)
        return read_value


def _get_cachedict() -> dict[str, _LRUFuncCache]:
    """Return the `cachedict` for the calling thread."""

    global _thread_local_data
    ca = getattr(_thread_local_data, "cachedict", None)
    if ca is None:
        _thread_local_data.cachedict = {}
    return _thread_local_data.cachedict


def _get_funccache(cached_func: Callable) -> _LRUFuncCache:
    """Return the cache for the `cached_func`.

    Raises:
        `linearmoney.exceptions.CacheError`:
            If there is no cache for `cached_func` in the current thread's `cachedict`.
    """

    _cachedict = _get_cachedict()
    funcname = cached_func.__qualname__
    try:
        return _cachedict[funcname]
    except KeyError:
        raise CacheError(f"Individual cache for: {funcname} not found.")


def invalidate(cached_func: Callable | None = None) -> None:
    """Invalidate the cache.

    Args:
        cached_func:
            If None (default), invalidate the entire cache for the calling thread.
            If not `None`, invalidate only the cache for the specified function in the
            calling thread.
    Raises:
        `linearmoney.exceptions.CacheError`:
            If the function passed as `cached_func` isn't cached.
            Individual caches are created dynamically, so this can happen if
            `cached_func` has not been called yet, but this is very unlikely in most
            applications.
    """

    if cached_func is None:
        for funccache in _get_cachedict().values():
            funccache.clear()
    else:
        _get_funccache(cached_func).clear()


_is_enabled = True


def enable(enable: bool) -> None:
    """Enable/disable package-wide caching."""

    if threading.current_thread() == threading.main_thread():
        global _is_enabled
        _is_enabled = enable
    else:
        global _thread_local_data
        _thread_local_data.is_enabled = enable


def is_enabled() -> bool:
    if threading.current_thread() == threading.main_thread():
        return _is_enabled
    else:
        global _thread_local_data
        ca = getattr(_thread_local_data, "is_enabled", None)
        if ca is None:
            _thread_local_data.is_enabled = _is_enabled
        return _thread_local_data.is_enabled


_base_size = 256


def get_base_size() -> int:
    """The current `base_size` for the cache.

    This value is used to calculate the dynamic `max_size` of each individual
    function cache. E.g. a function with 2.0 as its `size_multiplier` in the `cached`
    decorator will have `floor(2.0 * get_base_size())` maximum cache entries.
    """

    if threading.current_thread() == threading.main_thread():
        return _base_size
    else:
        global _thread_local_data
        ca = getattr(_thread_local_data, "base_size", None)
        if ca is None:
            _thread_local_data.base_size = _base_size
        return _thread_local_data.base_size


def set_base_size(new_base_size: int) -> None:
    """Set the current `base_size` for the cache.

    Raises:
        TypeError:
            If `new_base_size` doesn't support int.
    """

    if not isinstance(new_base_size, SupportsInt):
        raise TypeError(
            f"set_base_size(): Expected `SupportsInt`, \
got {type(new_base_size)}"
        )
    if threading.current_thread() == threading.main_thread():
        global _base_size
        _base_size = int(new_base_size)
    else:
        global _thread_local_data
        _thread_local_data.base_size = int(new_base_size)


def max_size(cached_func: Callable) -> int:
    """Return the maximum number of cache entries for the cache of the `cached_func`.

    Raises:
        `linearmoney.exceptions.CacheError`:
            If `cached_func` isn't cached.
            Individual caches are created dynamically, so this can happen if
            `cached_func` has not been called yet, but this is very unlikely in most
            applications.
    """

    return _get_funccache(cached_func).max_size


def size(cached_func: Callable | None = None) -> int:
    """The current size of the requested cache.

    Args:
        cached_func:
            If `None` (default) return the total combined size of all function caches,
            else return the total number of cached values for the cache of
            `cached_func`.
    Raises:
        `linearmoney.exceptions.CacheError`:
            If `cached_func` is not `None` and the function is not cached.
            Individual caches are created dynamically, so this can happen if
            `cached_func` has not been called yet, but this is very unlikely in most
            applications.
    """

    if cached_func is None:
        accumulator = 0
        for funccache in _get_cachedict().values():
            accumulator += funccache.size
        return accumulator
    else:
        return _get_funccache(cached_func).size


def head(cached_func: Callable) -> Any:
    """The cached value at the *head* (least recently used) position of the cache
    for `cached_func`.

    This is the value that will be removed from the cache if the `max_size` for
    `cached_func` is reached.

    Raises:
        `linearmoney.exceptions.CacheError`:
            If the passed function is not cached.
            Individual caches are created dynamically, so this can happen if
            `cached_func` has not been called yet, but this is very unlikely in most
            applications.
    """

    return _get_funccache(cached_func).head


def tail(cached_func: Callable) -> Any:
    """The cached value at the *tail* (most recently used) position of the cache
    for `cached_func`.

    Raises:
        `linearmoney.exceptions.CacheError`:
            If the passed function is not cached.
            Individual caches are created dynamically, so this can happen if
            `cached_func` has not been called yet, but this is very unlikely in most
            applications.
    """

    return _get_funccache(cached_func).tail


# Needed for type checking cache decorators.
T = TypeVar("T")
P = ParamSpec("P")


def _hit(
    func: Callable[P, T],
    *args,
    size_multiplier: int | float,
    **kwargs,
) -> T:
    """Hit the cache.

    If a call to func with equivalent arguments has already been made, return the
    cached value, else call `func`, cache the value, and return it.

    If `func` has not been called yet at all, create a new _LRUFuncCache for `func`
    and cache the value of the current call.

    The argument parsing portion of this function includes a check for numeric types
    like `decimal.Decimal` that makes sure that precision is taken into account when
    hashing arguments, so that for example, a rounding function that takes a decimal
    quantizer as an argument does not give the same result for decimals with different
    numbers of trailing zeros but the same actual value.
    """

    _funccache: _LRUFuncCache
    try:
        _funccache = _get_funccache(func)
    except CacheError:
        _funccache = _LRUFuncCache(
            funcname=func.__qualname__, size_multiplier=size_multiplier
        )
        _get_cachedict()[func.__qualname__] = _funccache
    key_accumulator: list[Hashable] = []
    if args:
        for i in args:
            if isinstance(i, Hashable):
                if isinstance(i, decimal.Decimal):
                    # Ensure Decimals of different precision are treated separately.
                    key_accumulator.append(str(i))
                else:
                    # Other supported numeric types need to be distinguished from
                    # unsupported numeric types with the same value.
                    # `id` is included to avoid natural hash collisions between numbers
                    # of the same type but different values. E.g. -1 and -2
                    key_accumulator.append((type(i), i, id(i)))
            else:
                key_accumulator.append(repr(i))
    if kwargs:
        for k, v in kwargs.items():
            if isinstance(v, Hashable):
                if isinstance(v, decimal.Decimal):
                    # Ensure Decimals of different precision are treated separately.
                    key_accumulator.append((k, str(v)))
                else:
                    # Other supported numeric types need to be distinguished from
                    # unsupported numeric types with the same value.
                    key_accumulator.append((k, type(v), v))
            else:
                key_accumulator.append((k, repr(v)))
    cache_key = tuple(key_accumulator)
    if not _funccache.is_cached(cache_key):
        value = func(*args, **kwargs)
        _funccache.write(cache_key, value)
        return value
    else:
        return _funccache.read(cache_key)


def cached(
    size_multiplier: int | float = 1,
) -> Callable[[Callable[P, T]], Callable[P, T]]:  # pragma: no cover
    """Used just like the `functools.lru_cache` decorator, but it allows unhashable
    types and has some special handling for numeric types, and in particular
    `decimal.Decimal` that takes precision into account, so that e.g. a rounding
    function that expects a decimal as an argument to be used in a
    `decimal.Decimal.quantize()` call will not treat two `decimal.Decimal`s with
    different trailing zeros as the same argument even if they have the same value."""

    def _outer_wrapper(func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(func)
        def _inner_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            if is_enabled():
                return _hit(
                    func,
                    size_multiplier=size_multiplier,
                    *args,
                    **kwargs,
                )
            else:
                return func(*args, **kwargs)

        return _inner_wrapper

    return _outer_wrapper



"""A library for working with money algebraically."""

from __future__ import annotations

__all__ = [
    "vector",
    "scalar",
    "data",
    "exceptions",
    "resources",
    "mixins",
    "cache",
]


import logging

# Setup logging
logger = logging.getLogger(__name__)

logger.info("linearmoney root logger initialized.")


from . import (  # noqa: E402
    vector,
    cache,
    data,
    exceptions,
    scalar,
    mixins,
    resources,
)

__version__ = "0.1.3"

CLDR_VERSION = resources.get_package_resource("cldr_version")
